#!/usr/bin/python3
from urllib.request import urlretrieve, urlopen
import os.path
import sys


_repositoryLocation = None
_packages = None

_packageCacheDirectory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'cache', 'package')
_repositoryCacheDirectory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'cache', 'repository')
_extractedCacheDirectory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'cache', 'extracted')
_extractedFilesDirectory = os.path.dirname(os.path.realpath(__file__))


def OpenRepository(location):
  from xml.etree.cElementTree import parse as xmlparse
  from time import time
  global _repositoryLocation, _packages
  _repositoryLocation = location
  # Check repository for latest primary.xml
  with urlopen(_repositoryLocation + 'repodata/repomd.xml') as metadata:
    doctree = xmlparse(metadata)
  xmlns = 'http://linux.duke.edu/metadata/repo'
  for element in doctree.findall('{%s}data'%xmlns):
    if element.get('type') == 'primary':
      primaryUrl = element.find('{%s}location'%xmlns).get('href')
  # Make sure all the cache directories exist
  try:
    for dir in [_packageCacheDirectory, _repositoryCacheDirectory, _extractedCacheDirectory]:
      os.makedirs(dir)
  except OSError: pass
  # Download repository metadata (only if not already in cache)
  primaryFilename = os.path.join(_repositoryCacheDirectory, os.path.splitext(os.path.basename(primaryUrl))[0])
  if not os.path.exists(primaryFilename):
    print('Dowloading repository data', file=sys.stderr)
    with urlopen(_repositoryLocation + primaryUrl) as primaryGzFile:
      import io, gzip
      primaryGzString = io.BytesIO(primaryGzFile.read())
      with gzip.GzipFile(fileobj=primaryGzString) as primaryGzipFile:
        with open(primaryFilename, 'wb') as primaryFile:
          primaryFile.writelines(primaryGzipFile)
  _packages = xmlparse(primaryFilename)


def _findPackage(packageName):
  packageBuildTime = 1272500000
  package = None
  xmlns = 'http://linux.duke.edu/metadata/common'
  if _packages:
    for element in _packages.findall('{%s}package'%xmlns):
      if element.find('{%s}name'%xmlns).text == packageName:
        if element.find('{%s}arch'%xmlns).text == 'noarch':
          elementBuildTime = int(element.find('{%s}time'%xmlns).get('build'))
          if elementBuildTime > packageBuildTime:
            package = element
            packageBuildTime = elementBuildTime
  return package


def _getPackageLocation(package):
  xmlns = 'http://linux.duke.edu/metadata/common'
  return package.find('{%s}location'%xmlns).get('href')


def _checkPackageRequirements(package, packageNames):
  xmlns = 'http://linux.duke.edu/metadata/common'
  rpmns = 'http://linux.duke.edu/metadata/rpm'
  allProviders = set()
  for req in package.findall('{%s}format/{%s}requires/{%s}entry'%(xmlns,rpmns,rpmns)):
    requirement = req.attrib['name']
    providers = set()
    for element in _packages.findall('{%s}package'%xmlns):
      for provides in element.findall('{%s}format/{%s}provides/{%s}entry'%(xmlns,rpmns,rpmns)):
        if provides.attrib['name'] == requirement:
          providers.add(element.find('{%s}name'%xmlns).text)
    if len(providers & packageNames) == 0:
      print('Package %s requires %s, provided by: %s' % (package.find('{%s}name'%xmlns).text,requirement,','.join(providers)), file=sys.stderr)
      allProviders.add(providers.copy().pop())
  return allProviders


def _downloadPackage(packageLocation):
  localFilename = os.path.basename(packageLocation)
  localFilenameFull = os.path.join(_packageCacheDirectory, localFilename)
  if not os.path.exists(localFilenameFull):
    print('Downloading', localFilename, file=sys.stderr)
    urlretrieve(_repositoryLocation + packageLocation, localFilenameFull)


def packagesDownload(packageNames, withDependencies=False):
  allPackageNames = {'mingw32-'+p for p in packageNames}
  packageNames = ['mingw32-'+p for p in packageNames]
  packageFilenames = []
  while len(packageNames) > 0:
    packName = packageNames.pop()
    package = _findPackage(packName)
    if package == None:
      print('Package %s not found' % packName, file=sys.stderr)
      continue
    dependencies = _checkPackageRequirements(package, allPackageNames)
    if withDependencies and len(dependencies) > 0:
      packageNames.extend(dependencies)
      allPackageNames |= dependencies
    packageLocation = _getPackageLocation(package)
    _downloadPackage(packageLocation)
    packageFilenames.append(os.path.basename(packageLocation))
  return packageFilenames


def _extractFile(filename, output_dir=_extractedCacheDirectory):
  from subprocess import check_call
  try:
    output = open('7z.log', 'w')
    check_call(['7z', 'x', '-o'+output_dir, '-y', filename], stdout=output)
    output.close()
    os.remove('7z.log')
  except:
    print('Failed to extract %s' % filename, file=sys.stderr)


def packagesExtract(packageFilenames):
  for packageFilename in packageFilenames :
    print('  Extracting', packageFilename, file=sys.stderr)
    cpioFilename = os.path.join(_extractedCacheDirectory, os.path.splitext(packageFilename)[0] + '.cpio')
    if not os.path.exists(cpioFilename):
      _extractFile(os.path.join(_packageCacheDirectory, packageFilename))
    _extractFile(cpioFilename, _extractedFilesDirectory)


def AddSuffix(packageNames, suffix):
  return packageNames | {p+'-'+suffix for p in packageNames if not p.endswith('-lang')}


def GetBaseDirectory():
  return os.path.join(_extractedFilesDirectory, 'usr/i686-w64-mingw32/sys-root/mingw')


def CleanExtracted():
  from shutil import rmtree
  rmtree(os.path.join(_extractedFilesDirectory, 'usr'), True)


def SetExecutableBit():
  # set executable bit on libraries and executables
  for root, dirs, files in os.walk(GetBaseDirectory()):
    for filename in files:
      if filename.endswith('.dll') or filename.endswith('.exe'):
        os.chmod(os.path.join(root, filename), 0o755)


def GetOptions():
  from optparse import OptionParser, OptionGroup

  parser = OptionParser(usage="usage: %prog [options] packages",
                        description="Easy download of RPM packages for Windows.")

  # Options specifiying download repository
  default_project = "windows:mingw:win32"
  default_repository = "openSUSE_11.4"
  default_repo_url = "http://download.opensuse.org/repositories/PROJECT/REPOSITORY/"
  repoOptions = OptionGroup(parser, "Specify download repository")
  repoOptions.add_option("-p", "--project", dest="project", default=default_project,
                         metavar="PROJECT", help="download from PROJECT,\ndefault: " + default_project)
  repoOptions.add_option("-r", "--repository", dest="repository", default=default_repository,
                         metavar="REPOSITORY", help="download from REPOSITORY,\ndefault: " + default_repository)
  repoOptions.add_option("-u", "--repo-url", dest="repo_url", default=default_repo_url,
                         metavar="URL", help="download packages from URL,\ndefault: " + default_repo_url)
  parser.add_option_group(repoOptions)

  # Other options
  parser.add_option("--no-clean", action="store_false", dest="clean", default=True,
                    help="do not remove previously extracted files before extracting packages")
  parser.add_option("--no-deps", action="store_false", dest="withdeps", default=True,
                    help="do not download dependencies")
  parser.add_option("-z", "--make-zip", action="store_true", dest="makezip", default=False,
                    help="Make a zip file of the extracted packages.  The name of the zip file is based on the first package specified.")
  parser.add_option("-q", "--quiet", action="store_false", dest="verbose", default=True,
                    help="don't print status messages to stdout")

  (options, args) = parser.parse_args()

  if len(args) == 0:
    parser.print_help(file=sys.stderr)
    sys.exit(1)

  return (options, args)


def main():
  (options, args) = GetOptions()
  packages = set(args)

  # Open repository
  repository = options.repo_url.replace("PROJECT", options.project.replace(':', ':/')).replace("REPOSITORY", options.repository)
  try:
    OpenRepository(repository)
  except Exception as e:
    sys.exit('Error opening repository:\n\t%s\n\t%s' % (repository, e))

  if options.clean:
    CleanExtracted()

  if options.makezip:
    package = _findPackage("mingw32-"+args[0])
    if package != None:
      zipfileName = os.path.basename(_getPackageLocation(package))
      zipfileName = zipfileName.replace('mingw32-', '').replace('.rpm', '').replace('.noarch', '') + '.zip'

  packages = packagesDownload(packages, options.withdeps)
  for package in sorted(packages):
    print(package)

  packagesExtract(packages)
  SetExecutableBit()

  if options.makezip:
    from zipfile import ZipFile
    packagezip = ZipFile(zipfileName, 'w')
    for root, dirs, files in os.walk(GetBaseDirectory()):
      for filename in files:
        fullname = os.path.join(root, filename)
        packagezip.write(fullname, fullname.replace(GetBaseDirectory(), ''))
    if options.clean:
      CleanExtracted()

if __name__ == "__main__":
    main()

